
# The opposite of the cloud
## &nbsp;
## Tom Eastman
## &nbsp;
<img style="box-shadow: none; border: 0; " width="50%" src="koordinates_logo.svg">

-

Good morning everyone!

I'm Tom. I've worn a couple hats over the last few years. I guess I'm first and foremost a Python developer, but I've done time both as a sysadmin and as a security consultant. These days I'm back in the DevOps-y sys-admin-y space.

Just to be clear: this isn't a keynote.

This is a talk about the project I'm working on, where I get to dive elbow deep into a bunch of interesting technologies, in the hope of solving what seems to me to be a fairly novel problem.

It's been a lot of fun for me, and I hope you'll find it interesting.

...

I'm here to talk about the 'opposite' of the cloud.

So FIRST OF ALL, what is the cloud?

---

<img style="box-shadow: none; border: 0;" width="50%"  src="there-is-no-cloud.png">

-

The cloud is "somebody else's computer".

That's an old joke, but whether it's *funny* or not depends on who you're talking to.

Using the cloud means that you're running critical parts of your own infrastructure on hardware that isn't under your direct control.

---

<img style="box-shadow: none; border: 0; " width="90%" src="kx-screenshot.png">

-

Second: umm, what is koordinates?

Koordinates is a cloud-based data platform for processing and publishing large, complex spatial datasets.

The primary instance of the koordinates platform lives in the cloud (on aws), but we also have a presence on client networks.

The internal presence is necessary because of the sheer amount of information involved with geospatial data publishing.

I came on board with the goal of building an appliance that would make this easier for both koordinates and their customers to handle.

---

### An... appliance?

-

The goal is for there to be a secure and easy way to link internal resources to the cloud-based data publishing platform, for customers who don't have the developer resources to write their own using Koordinates' APIs.

The appliance will, for example, link a clients internal 'staging' server, which might just be a CIFS server, or ARCGIS, or something, and make it easy to synchronize new or updated data on there to the published version of the data on koordinates.com

Similarly, a customer might have an interest in keeping something that lives on the koordinates platform synchronised to, say, a  local postgresql or POSTGIS database, the Koordinates Data Gateway (that's what I'm talking about) will be the bridge through which we can push updates to their local resource immediately when there's an update upstream

---

## Requirements

* A bootable .ova, or AMI/cloud images. <!-- .element: class="fragment" -->
* Single-step configuration <!-- .element: class="fragment" -->
* Subsequent configuration via website. <!-- .element: class="fragment" -->
* Simple firewall requirements. <!-- .element: class="fragment" -->

-

So what are the basic requirements of an appliance like this?

[click] Machine images ready for whatever virtualization environment the customer is comfortable with. Including cloud-ready ones if the customer's internal network isn't as internal as expected.

[click] The configuration on the appliance has to be as close to zero effort as possible. Preferably, just one step, like, typing in an activation code

[click] After the initial step of configuring the appliance, all subsequent configuration happens on the koordinates website, which we then push back down to the appliance.

[click] The network and firewall requirements of the appliance should be as simple as possible for the customer's network administrators. That doesn't mean "permissive", just that it's simple to give the appliance the access that it needs, while also locking it down as much they feel comfortable with.

---

## Do these requirements sound familiar?

-

Does this set of requirements sound familiar to us?

---

## This sounds like an 'Internet of Things' device.

-

It sounds an awful lot like an Internet of Things device, doesn't it?

Something that runs inside your trusted network, that you administer via a web interface somewhere else.

---

## This is the opposite of the cloud.

-

And so maybe I'm stretching the metaphor in order to get a catchy title, but: If "the cloud" is running your infrastructure on "Somebody else's computer", then the opposite of the cloud is letting someone else run their infrastructure on one of *your* computers.

My job is to try to find a way of building something like this so that, even the most jaded, perenially angry system administrator can grudgingly say: "well, okay, I guess that sounds like it'll work."

I might be a glutten for punishment.

---

## What do IoT devices get wrong?

- Security patches/firmware updates/software updates don't happen. <!-- .element: class="fragment" -->
- Treating the network as non-hostile. <!-- .element: class="fragment" -->
- Difficult to remotely administer/maintain. <!-- .element: class="fragment" -->

-

What is it that IoT devices get wrong that have given them a poor reputation?

---

### How do you make a secure IoT device?

- no default or static credentials <!-- .element: class="fragment" -->
- reduce its attack surface <!-- .element: class="fragment" -->
- secure all network comms  <!-- .element: class="fragment" -->
- ensure that it fails securely <!-- .element: class="fragment" -->
- make sure it's easy to update <!-- .element: class="fragment" -->

-

1. Make sure there are no default or static credentials of any sort.
2. Reduce its attack surface to next-to-zero by strictly firewalling incoming and outgoing communications.
3. Make sure all network communications are secure (as in: authenticated and inscrutable)
4. Ensure that it fails *securely*. For example, if it's been unable to connect to the orchestration server, or check for updates, cut off its network access.
5. Make sure you can push updates to it without delay.

---

### Solution: <br> Don't treat appliances like appliances.

### Treat them like tightly orchestrated linux servers. <!-- .element: class="fragment" -->

-

So at first I was looking at appliance-centric scaled down distributions. But the problem is that they have a different set of priorities than me. They generally want to cut disk space/image size, reduce memory requirements, and pack everything down.

Whereas, my top priority is to be able to apply security updates, be able to apply security updates, and be able to apply security updates

The answer is that you just can't treat your appliance like a thing that is built, and then ignored.

Instead, you treat them like what they need to be: tightly orchestrated linux servers themselves.

[click]

---

### Sticking to a conservative architecture

- a mainstream linux distribution, minimally customized <!-- .element: class="fragment" -->
- a strong server orchestration and management framework <!-- .element: class="fragment" -->

-

Using a Linux Distribution with a solid community behind it makes every single kind of sense:

 - I have access to the community, and the package repository, and the documentation
 - Distribution provided security updates, applied with unattended-upgrades and auto-reboot
 - Using Debian, thanks to DKMS I even get kernel security updates with magical auto-compilation of third-party kernel modules, such as the hypervisor guest additions for VMWare and Virtualbox

---

## The solution components.

- aspen: a customized Debian machine image built with Packer <!-- .element: class="fragment" -->
- pando: orchestration server/C&C network <!-- .element: class="fragment" -->
- hakea: a Django/REST microservice API in charge <!-- .element: class="fragment" -->

-

Aspen is the appliance itself, and what's running on it.

Pando is the network and the server on the koordinates side.

Hakea is the web API that controls Pando.

Over the next few slides, I'm going to go through a bunch of the technologies that all come together to make this work.

---

## Saltstack command and control

- "a configuration management system, capable of maintaining remote nodes in defined states" <!-- .element: class="fragment" -->

- "a distributed remote execution system used to execute commands and query data on remote nodes" <!-- .element: class="fragment" -->

-

Using a strong orchestration framework means you can keep all appliances up-to-date with changes to your processes, and issue ad-hoc commands easily.

Saltstack is a master/slave server orchestration and management tool, like puppet, and like Ansible Tower.

Quoting from the 'Introduction to Salt' page.

[read]

Every salt minion maintains a persistent outgoing TCP connection with the server, it waits to receive orchestration commands or execution commands.

For this project, salt becomes the command and control system, and the configuration orchestration server, all in one.

Appliances can be controlled and kept up-to-date with a tool designed for the job.

The appliance is 'disposable', if something goes wrong you can kill it, boot up another one, and it is reconfigured to its old state by saltstack.

---

## OpenVPN as Internet transport

-

In order to provide some guarantees around the security of the communications between the appliance and the orchestration server, all its communications are over an OpenVPN network.

Because of this, we can be super strict with both incoming and outgoing firewalls. We can restrict all outgoing network traffic to just a handful of things, and we can restrict *internet* traffic to only two things: TCP port 443 to our activation endpoint, and UDP port 1194 to our OpenVPN server.

In practice, the appliance might have to use a number of internet services, such as Debian package repositories, time servers for clock synchronization, remote logging, and of course the saltstack connection mentioned before. All these services are connected through the VPN to services we control on the cloud side.

---

## So what is the appliance, then?

- A lightly mangled Debian Jessie VM image. <!-- .element: class="fragment" -->
- Boots into a simple configuration screen. <!-- .element: class="fragment" -->

-

So, what is the appliance, what's it built out of, and what's going to do?

Well, it's pretty much a debian box, with a bunch of packages and scripts installed on it.

It's built using Packer.io, which gives me a standard build environment for all my target platforms -- virtualbox, vmware, aws, openstack, etc.

It feels like cross-compiling a binary, except the binary is a whole computer.

---

<img style="box-shadow: none; border: 0;" src="configure-data-gateway.png">

---

## The appliance (cont.)

- connects to the Saltstack server over the VPN <!-- .element: class="fragment" -->
- waits to be told what to do <!-- .element: class="fragment" -->
- application services are deployed as docker containers  <!-- .element: class="fragment" -->

-

So, kind of of note here is that, everything I've done so far is entirely generic. Nothign at all related to the actual services that Koordinates might actually be *offering* through these appliances.

Those actual services will be implemented in Docker containers that are deployed and executed on the appliances. This turns out to be a pretty great way to ensure loose coupling between the architecture of the network I'm building, and the actual business logic of the service.

---

## Appliance authentication

- Activation request sent (code + machine id)  <!-- .element: class="fragment" -->
- VPN and Salt certs returned in response <!-- .element: class="fragment" -->
- Establish link to server through VPN <!-- .element: class="fragment" -->

-

So, you can't just download one of these appliance images and then be on my network. You gotta pass the test first.

An appliance that hasn't been activated is inert, it doesn't have a VPN connection and the strict firewall rules mean it can't make much in the way of network connections. It'll just sit there using up cycles and a DHCP-provided IP address.

When you configure the appliance by typing in your activation code, a couple things happen.

[click]

A request is sent over https to the onboarding endpoint on Hakea, the Django microservice API I mentioned before.

[click]

If the request is valid, then a VPN certificate and a Salt minion certificate, are generated and delivered to the appliance. The appliance puts them in place, establishes the VPN link, and then kicks off the Salt minion service.

This was a chicken and egg problem that went through several iterations

[elaborate]

---

## Security through self-destruct

- Short-lived VPN certificates that have to be replaced. <!-- .element: class="fragment" -->
- Keys are rotated on a daily schedule <!-- .element: class="fragment" -->
- Without the keys, the appliance drops off the net <!-- .element: class="fragment" -->

-

An un-patched device that can talk to the internet is a disaster waiting to happen.

So, how do you make sure that the devices that you're responsible for are always up-to-date?

By making sure that, if you lose the ability to update the appliance, it won't stay on the Internet very long.

[click]

I think people are pretty used to long-lived VPN certificates, but there's a lot to be said for having VPN credentials that don't last long.

My goal is to make sure that, if I lose the ability to issue commands to an appliance, it doesn't stay online to potentially cause any problems.

So the VPN certs are only valid for a day, and if I can't talk to the appliance, it doesn't get a new one. Without a new cert, it can't talk to the internet.

I haven't yet, but probably will, add a check that just shuts down the appliance if it hasn't been able to re-establish new VPN credentials after some time period. I like the idea of the appliances just dying themselves without a fuss.

---

## Hakea: REST control

- a Django REST Framework microservice <!-- .element: class="fragment" -->
- RESTful resources for appliances and appliance configuration <!-- .element: class="fragment" -->
- self-documenting using DRF and CoreAPI Schema <!-- .element: class="fragment" -->

-

So, we have a network of disparate virtual machines running on different customer's networks, communicating via an encrypted channel to an endpoint in our infrastructure, connected to our Salt server and waiting to take commands.



---

## Hakea (contd.)

- exposes the appliance activation API endpoint
- exposes a REST interface to the kx platform
- exposes a Salt 'pillar' data source to Saltstack
- sends execution commands to appliances via 'salt-api'
- communicates with Hashicorp Vault, which handles PKI

---

## DevOps principles apply beyond the cloud

-

This is a young project, I've learned a lot in a short period of time and there are a lot of interesting challenges still to go.

But this is the point I want to make: Using careful, disciplined DevOps principles makes a hard problem like this actually pretty feasible.

Keeping the components loosely coupled, easily re-buildable, testable, and *disposable* means you can do cool things you couldn't do before.

---

## Thank you for your time!
